from __future__ import annotations

import curses
import curses.ascii
import html
import os
import shutil
import subprocess
import sys
import webbrowser
from threading import Event, RLock, Thread

from . import __version__
from .curses_utils import App, win_addstr, win_help
from .curses_utils.list1m import List1m, ListProto1m
from .db import Favourites, Record, from_url
from .utils import Mpv, RadioException, ThreadStr, search_lyrics, search_words_url, str2clipboard

HELP = [
    ("h", "This help screen"),
    ("q, Esc", "Quit the program"),
    ("j, Down", "Move selection down"),
    ("k, Up", "Move selection up"),
    ("PgUp", "Page up"),
    ("PgDown", "Page down"),
    ("g, Home", "Move to first item"),
    ("G, End", "Move to last item"),
    ("Right, Left", "Move to/from a directory"),
    ("Shift-{Up,Down}", "Move a record up/down"),
    ("Insert", "Add to Favourites"),
    ("Delete", "Delete from Favourites"),
    ("Enter", "Play selected radio"),
    ("Space", "Stop/Resume"),
    ("Ctrl-U", "Copy URL to clipboard"),
    ("Ctrl-L", "Copy lyrics to clipboard"),
    ("L", "Search lyrics in browser"),
]

OPML_URL = 'https://opml.radiotime.com/'


class Main(App, ListProto1m):  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    def __init__(self, screen):
        super().__init__(screen)

        curses.init_pair(1, curses.COLOR_CYAN, curses.COLOR_BLACK)

        self.record = Record({})  # root record
        self.fav = Favourites(self.record)
        self.fav.load_from_home()
        self.record.children.append(self.fav)

        self.mpv = Mpv()
        self.radio = ''

        self.song = ThreadStr(RLock())
        self.status_lock = RLock()
        self.thread_meta = None
        self.stop_meta = Event()
        self.status_str = ''

        self.create_windows()

    def from_url(self, url: str, r: Record) -> bool:
        try:
            from_url(url, r)
            return True
        except RadioException as e:
            self.status(f'Error: {e}')
            return False

    def load_opml(self):
        self.status(f'Load OPML: {OPML_URL} ...')
        if self.from_url(OPML_URL, self.record):
            self.status('OK')
        self.refresh_all()

    def create_windows(self):
        '''
        radio-curses v...

        records ...
        ------
        status
        '''
        maxy, maxx = self.screen_size

        rows, cols = (maxy - 4, maxx - 2)

        win = self.screen.derwin(rows, cols, 2, 1)
        self.win = List1m(win, self, current_color=curses.color_pair(1) | curses.A_BOLD)

        # status
        self.win3 = self.screen.derwin(1, maxx, maxy - 1, 0)

    def refresh_win_deps(self):
        pass

    def get_record(self, i: int) -> Record | None:
        len_ = len(self.record)
        if not i < len_:
            return None
        return self.record[i]

    def get_record_str(self, i: int) -> str:
        if not (r := self.get_record(i)):
            return ''
        if r.isdir():
            return f'{r.text}/'
        return r.text

    def records_len(self) -> int:
        return len(self.record)

    def refresh_all(self):
        self.screen.erase()

        s = f'radio-curses v{__version__} (h - Help)'
        proxy = os.environ.get('http_proxy') or os.environ.get('https_proxy')
        if proxy:
            s += f' {proxy=}'
        win_addstr(self.screen, 0, 1, s)
        self.screen.refresh()

        maxy, maxx = self.screen_size
        win = self.screen.derwin(maxy - 2, maxx, 1, 0)
        win.erase()
        win.box()
        win.refresh()

        self.win.refresh()

        self.status(self.status_str, force=True)

    def right(self, i: int):
        if not (r := self.get_record(i)):
            return
        self.record.pos = (self.win.cur, self.win.idx)
        if r.isdir():
            self.record = r
            if not self.record.children and 'URL' in r.d:
                self.from_url(r.d['URL'], self.record)
            self.win.cur, self.win.idx = self.record.pos
            self.win.refresh()

    def left(self):
        if not self.record.parent:
            return
        self.record.pos = (self.win.cur, self.win.idx)
        self.record = self.record.parent
        self.win.cur, self.win.idx = self.record.pos
        self.win.refresh()

    def poll_metadata(self, stop: Event):
        while True:
            if stop.wait(5):
                return
            d = self.mpv.get_metadata(stop)
            d2 = d.get('data')
            if not d2:
                continue
            song = d2.get('icy-title')
            if song:
                song = song.rstrip()
                song = html.unescape(song)
                self.song.set(song)
                s = f'{self.radio}: {song}'
            else:
                s = f'{self.radio}'
            self.status(s)

    def start_player(self, i: int):
        if not (r := self.get_record(i)):
            return
        self.record.pos = (self.win.cur, self.win.idx)
        if r.isaudio():
            self.radio = r.text
            self.status(f'Start {self.radio} ...')
            self.mpv.start(r.d['URL'])
            if not self.thread_meta:
                self.thread_meta = Thread(target=self.poll_metadata, args=(self.stop_meta,))
                self.thread_meta.start()

    def record_move_up(self, i: int) -> bool:
        return self.record.move_child_up(i)

    def record_move_down(self, i: int) -> bool:
        return self.record.move_child_down(i)

    def add_to_favourites(self, i: int):
        if not (r := self.get_record(i)):
            return
        if self.fav.add_record(r):
            self.status('Added to Favourites')

    def del_favourite(self, i: int):
        if self.record != self.fav or not self.get_record(i):
            return
        del self.fav[i]
        self.win.refresh()

    def url2clipboard(self):
        if not (r := self.get_record(self.win.idx)):
            return
        if url := r.d.get('URL', None):
            str2clipboard(url)
            self.status(f'URL copied: {url}')
        else:
            self.status('URL is empty')

    def show_lyrics(self):
        song = self.song.get()
        if not song:
            return
        lyrics = search_lyrics(song)
        if lyrics:
            str2clipboard(lyrics)
            self.status('lyrics copied to clipboard')

    def search_lyrics(self):
        song = self.song.get()
        if not song:
            return
        if not (url := search_words_url(song)):
            return
        try:
            if shutil.which("xdg-open"):
                subprocess.run(['xdg-open', url], check=False)
            else:
                webbrowser.open(url)
        except ImportError:
            self.status(f'Could not load python module "webbrowser" for {url=}')

    def status(self, s: str, force=False):
        if curses.isendwin():
            return
        if force or self.status_str != s:
            with self.status_lock:
                self.win3.erase()
                win_addstr(self.win3, 0, 1, s)
                self.win3.refresh()
                self.status_str = s

    def shutdown(self, *_):
        self.status('Closing ...')
        self.fav.save_to_home()
        if self.thread_meta and self.thread_meta.is_alive():
            self.stop_meta.set()
            self.thread_meta.join()
        self.mpv.stop()
        sys.exit(0)

    def run(self):
        self.refresh_all()
        self.load_opml()
        self.input_loop()

    def input_loop(self):  # pylint: disable=too-many-branches,too-many-statements
        for char_ord in self.getch():
            char = chr(char_ord)

            if self.win.handle_input(char_ord):
                pass
            elif char_ord == curses.KEY_RIGHT:
                self.right(self.win.idx)
            elif char_ord == curses.KEY_LEFT:
                self.left()
            elif char_ord == curses.ascii.LF:  # Enter
                self.start_player(self.win.idx)
            elif char_ord == curses.ascii.SP:  # Space
                rt = self.mpv.toggle()
                if rt > 0:
                    self.status(f'Start {self.radio} ...')
                elif rt < 0:
                    self.status(f'Stop {self.radio} ...')
            elif char_ord == curses.KEY_IC:  # insert
                self.add_to_favourites(self.win.idx)
            elif char_ord == curses.KEY_DC:  # delete
                self.del_favourite(self.win.idx)
            elif char.upper() == 'H':  # Print help screen
                win_help(self.screen, HELP)
                self.refresh_all()
            elif char == 'L':
                self.search_lyrics()
            elif char_ord == 21:  # ^U
                self.url2clipboard()
            elif char_ord == 12:  # ^L
                self.show_lyrics()


def main2(screen):
    app = Main(screen)
    app.run()


def main():
    curses.wrapper(main2)


if __name__ == '__main__':
    main()



import curses
import curses.ascii
import os
import sys
from contextlib import contextmanager
from signal import SIGINT, SIGTERM, SIGWINCH, signal
from typing import Generator


def win_addstr(
    win: curses.window, row: int, col: int, s: str, attr: int = 0, border: int = 0, align: int = -1
):  # pylint: disable=too-many-arguments,too-many-positional-arguments
    'align: -1 (left), 0(center), 1(right)'
    try:
        _, cols = win.getmaxyx()
        cols2 = cols - col - border
        s = s[:cols2]
        if align == 0:
            s = f'{s:^{cols2}}'
        elif align > 0:
            s = f'{s:>{cols2}}'
        win.addstr(row, col, s, attr)
    except curses.error:
        # https://docs.python.org/3/library/curses.html#curses.window.addstr
        # Attempting to write to the lower right corner of a window, subwindow, or pad
        # will cause an exception to be raised after the string is printed.
        pass


def win_center(screen: curses.window, rows: int, cols: int, header: str, color: int = 0) -> curses.window:
    max_rows, max_cols = screen.getmaxyx()
    rows = min(rows, max_rows)
    cols = min(cols, max_cols)
    y = (max_rows - rows) // 2
    x = (max_cols - cols) // 2

    win = screen.derwin(rows, cols, y, x)
    win.keypad(True)
    win.erase()
    if color:
        win.attrset(color)
    win.box()

    header = header[:cols]
    x = (cols - len(header)) // 2
    win_addstr(win, 0, x, header)

    return win


def ask_delete(screen: curses.window, color: int = 0) -> bool:
    header = 'Delete current record'
    win = win_center(screen, 5, 30, header, color=color)

    win_addstr(win, 1, 1, 'Are you sure?', border=1, align=0)
    win_addstr(win, 3, 1, 'Press Y to delete ...', border=1, align=0)

    try:
        ch = win.getch()
        if ch == ord('Y'):
            return True
        return False
    except curses.error:
        return False
    finally:
        # https://stackoverflow.com/questions/2575409/how-do-i-delete-a-curse-window-in-python-and-restore-background-window
        win.erase()
        del win
        screen.touchwin()


def win_help(screen, help_: list[tuple[str, str]]):  # pylint: disable=too-many-locals
    '''
    help_ = [
        (key1, help1),
        (key2, help2),
        ...
    ]
    '''
    header = "Help information:"
    footer = "Press any key to continue..."

    lmax = max(len(i[0]) for i in help_)  # (lmax) - help

    def iter_help():
        for i, j in help_:
            yield f'{i:<{lmax}} - {j}'  # keys - help

    rows = len(help_) + 1  # footer=1
    rows2 = rows + 2  # border=2
    cols = max(len(header), max(len(i) for i in iter_help()), len(footer))
    cols2 = cols + 2  # border=2

    win = win_center(screen, rows2, cols2, header)
    rows2, cols2 = win.getmaxyx()

    row = 0
    col = 1
    for s in iter_help():
        row += 1
        win_addstr(win, row, col, s, border=1)
    row += 1
    win_addstr(win, row, col, footer, border=1)

    # Wait for any key press
    win.getch()

    # https://stackoverflow.com/questions/2575409/how-do-i-delete-a-curse-window-in-python-and-restore-background-window
    win.erase()
    del win
    screen.touchwin()


class App:
    def __init__(self, screen):
        self.screen = screen

        self.orig_sigint = signal(SIGINT, self.shutdown)
        signal(SIGTERM, self.shutdown)
        signal(SIGWINCH, self.sigwinch_handler)

        self.screen.keypad(1)
        curses.curs_set(0)
        curses.noecho()
        curses.start_color()

        self.screen_size = (curses.LINES, curses.COLS)  # pylint: disable=no-member

    def sigwinch_handler(self, *_):
        maxx, maxy = os.get_terminal_size()
        self.screen_size = (maxy, maxx)
        self.create_windows()
        self.refresh_all()

    def shutdown(self, *_):
        if curses.isendwin():
            # return to curses
            self.screen.refresh()
        sys.exit(0)

    def create_windows(self):
        pass

    def refresh_all(self):
        pass

    def handle_alt_key(self, ch: int):
        pass

    def _handle_alt_key(self):
        # https://stackoverflow.com/a/22362849
        ch = -1
        try:
            self.screen.nodelay(True)
            ch = self.screen.getch()  # get the key pressed after ALT
        finally:
            self.screen.nodelay(False)
        if ch == -1:
            self.shutdown()
        self.handle_alt_key(ch)

    def getch(self) -> Generator[int, None, None]:
        while True:
            try:
                ch = self.screen.getch()

                if ch == -1:
                    # SIGWINCH interrupt
                    # t = self.screen.getmaxyx()  # doesn't work
                    continue

                if ch == curses.ascii.ESC:  # Esc
                    self._handle_alt_key()
                elif ch == curses.KEY_RESIZE:
                    # doesn't work
                    # t = self.screen.getmaxyx()
                    pass
                else:
                    char = chr(ch)
                    if char.upper() == 'Q':
                        self.shutdown()
                    else:
                        yield ch
            except curses.error:
                pass


@contextmanager
def escape2terminal(app: App):
    def _clear():
        # clear terminal for privacy
        os.system('cls' if os.name == 'nt' else 'clear')

    curses.endwin()
    old = signal(SIGINT, app.orig_sigint)
    try:
        _clear()
        yield
    except KeyboardInterrupt:
        pass
    finally:
        _clear()
        app.screen.refresh()
        signal(SIGINT, old)


def input_search(app: App, prompt: str) -> tuple[bool, str]:  # ok, search str
    with escape2terminal(app):
        return (True, input(prompt))
    return (False, '')



from __future__ import annotations

import json
import os
import shutil
import socket
import subprocess
from contextlib import contextmanager
from subprocess import PIPE, Popen
from threading import Event, RLock


class RadioException(Exception):
    pass


@contextmanager
def unix_socket(timeout: float = 2.0):
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.settimeout(timeout)
    try:
        yield s
    finally:
        s.close()


class Mpv:
    def __init__(self):
        self.socket = os.path.expanduser('~/tmp/mpv.socket')
        mpv = shutil.which('mpv') or '/usr/bin/mpv'
        self.cmd = [mpv, '--terminal=no', f'--input-ipc-server={self.socket}']
        self.proc = None
        self.url = None

    def stop(self):
        if self.proc is not None:
            self.proc.terminate()
            self.proc.wait()
        if os.path.exists(self.socket):
            os.remove(self.socket)
        self.proc = None

    def start(self, url: str):
        if not url:
            return
        self.stop()
        if url.startswith('http://'):
            url = f'{url[:4]}s{url[4:]}'  # https://
        cmd: list[str] = self.cmd + [url]
        self.proc = subprocess.Popen(  # pylint: disable=consider-using-with
            cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL
        )
        self.url = url

    def toggle(self) -> int:
        # -1: stop, 0: nothing, 1: start
        rt = 0
        if self.proc:
            self.stop()
            rt = -1
        elif self.url:
            self.start(self.url)
            rt = 1
        return rt

    def send_command(self, cmd: dict, stop: Event) -> dict:
        with unix_socket() as client:
            while True:
                if not os.path.exists(self.socket):
                    if stop.wait(2):
                        return {}
                    continue
                try:
                    client.connect(self.socket)
                    break
                except (ConnectionError, socket.timeout):
                    if stop.wait(2):
                        return {}
                    continue
            bytes_ = json.dumps(cmd).encode('utf-8') + b'\n'
            try:
                client.sendall(bytes_)
                return socket2json(client)
            except (socket.timeout, UnicodeDecodeError):
                return {}

    def get_metadata(self, stop: Event) -> dict:
        return self.send_command({'command': ['get_property', 'metadata']}, stop)


def socket2json(s: socket.socket) -> dict:
    try:
        with s.makefile(encoding='utf-8', errors='replace') as fp:
            return json.loads(fp.readline())
    except (ConnectionResetError, json.JSONDecodeError):
        return {}


def str2clipboard(s: str):
    with Popen(['xsel', '-b', '-i'], stdout=PIPE, stdin=PIPE, stderr=PIPE, text=True) as p:
        p.communicate(input=s)


class ThreadStr:
    'str between threads'

    def __init__(self, lock: RLock):
        self.lock = lock
        self.value = ''

    def get(self) -> str:
        with self.lock:
            return self.value

    def set(self, value: str) -> bool:
        with self.lock:
            if self.value != value:
                self.value = value
                return True
        return False


# LYRICS_SEARCH_URL = 'https://lrclib.net/search/'
LYRICS_SEARCH_URL = 'https://duckduckgo.com/?q=lyrics '
DEL_PREFIXES = ('Now Playing: ',)
REPLACE2SPACE = ('_', '/', '-', ':', '+', '(', ')', '!')


def search_words(song: str) -> list[str] | None:
    if song:
        for i in DEL_PREFIXES:
            song = song.removeprefix(i)
        for i in REPLACE2SPACE:
            song = song.replace(i, ' ')
        return song.split()
    return None


def search_words_url(song: str) -> str | None:
    if not (words := search_words(song)):
        return None
    s = ' '.join(words)
    return f'{LYRICS_SEARCH_URL}{s}'



from __future__ import annotations

import json
import os
import shutil
import socket
import subprocess
from contextlib import contextmanager
from subprocess import PIPE, Popen
from threading import Event, RLock

import requests
from requests.exceptions import ReadTimeout

from . import __homepage__, __project_name__, __version__


class RadioException(Exception):
    pass


@contextmanager
def unix_socket(timeout: float = 2.0):
    s = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    s.settimeout(timeout)
    try:
        yield s
    finally:
        s.close()


class Mpv:
    def __init__(self):
        self.socket = os.path.expanduser('~/tmp/mpv.socket')
        mpv = shutil.which('mpv') or '/usr/bin/mpv'
        self.cmd = [mpv, '--terminal=no', f'--input-ipc-server={self.socket}']
        self.proc = None
        self.url = None

    def stop(self):
        if self.proc is not None:
            self.proc.terminate()
            self.proc.wait()
        if os.path.exists(self.socket):
            os.remove(self.socket)
        self.proc = None

    def start(self, url: str):
        if not url:
            return
        self.stop()
        if url.startswith('http://'):
            url = f'{url[:4]}s{url[4:]}'  # https://
        cmd: list[str] = self.cmd + [url]
        self.proc = subprocess.Popen(  # pylint: disable=consider-using-with
            cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL, stdin=subprocess.DEVNULL
        )
        self.url = url

    def toggle(self) -> int:
        # -1: stop, 0: nothing, 1: start
        rt = 0
        if self.proc:
            self.stop()
            rt = -1
        elif self.url:
            self.start(self.url)
            rt = 1
        return rt

    def send_command(self, cmd: dict, stop: Event) -> dict:
        with unix_socket() as client:
            while True:
                if not os.path.exists(self.socket):
                    if stop.wait(2):
                        return {}
                    continue
                try:
                    client.connect(self.socket)
                    break
                except (ConnectionError, socket.timeout):
                    if stop.wait(2):
                        return {}
                    continue
            bytes_ = json.dumps(cmd).encode('utf-8') + b'\n'
            try:
                client.sendall(bytes_)
                return socket2json(client)
            except (socket.timeout, UnicodeDecodeError):
                return {}

    def get_metadata(self, stop: Event) -> dict:
        return self.send_command({'command': ['get_property', 'metadata']}, stop)


def socket2json(s: socket.socket) -> dict:
    try:
        with s.makefile(encoding='utf-8', errors='replace') as fp:
            return json.loads(fp.readline())
    except (ConnectionResetError, json.JSONDecodeError):
        return {}


def str2clipboard(s: str):
    with Popen(['xsel', '-b', '-i'], stdout=PIPE, stdin=PIPE, stderr=PIPE, text=True) as p:
        p.communicate(input=s)


class ThreadStr:
    'str between threads'

    def __init__(self, lock: RLock):
        self.lock = lock
        self.value = ''

    def get(self) -> str:
        with self.lock:
            return self.value

    def set(self, value: str) -> bool:
        with self.lock:
            if self.value != value:
                self.value = value
                return True
        return False


LYRICS_URL = 'https://lrclib.net/api/search'  # /api/search?q=still+alive+portal
DEL_PREFIXES = ('Now Playing: ',)
REPLACE2SPACE = ('_', '/', '-', ':', '+', '(', ')', '!', 'feat.')
USER_AGENT = f'{__project_name__} v{__version__} ({__homepage__})'


def search_lyrics(song: str) -> str | None:
    if song:
        for i in DEL_PREFIXES:
            song = song.removeprefix(i)
        for i in REPLACE2SPACE:
            song = song.replace(i, ' ')
        search = '+'.join(song.split())
        if search:
            try:
                resp = requests.get(
                    LYRICS_URL,
                    headers={'User-Agent:': USER_AGENT},
                    params={'q': search},
                    timeout=5,
                )
            except ReadTimeout:
                return None
            resp.raise_for_status()
            l: list[dict] = resp.json()
            if l:
                d = l[0]
                if lyrics := d.get('plainLyrics'):
                    return lyrics
    return None
